==> wikitty-api-2.1-sources.jar: org/nuiton/wikitty/WikittyCopyOnWrite.java <==
package org.nuiton.wikitty;

import java.beans.PropertyChangeListener;
import java.math.BigDecimal;
import java.util.Collection;
import java.util.Date;
import java.util.List;
import java.util.Set;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/** this class wrap a wikitty in a proxy that copy the wikitty before applying any change
 *
 * It is used for caching purpose. This class own a reference to an actual
 * wikitty, multiple instance of this class can own some references to a same
 * actual wikitty. To prevent the target to be modified, a copy is created
 * to prevent side-effect.
 * 
 * So, when a wikitty restored from cache is modified, a copy is modified, so
 * if change are cancelled, next restore will restore the original and not
 * the modified version (until the modified version is stored).
 * 
 * used in {@link WikittyServiceCached}
 */
public class WikittyCopyOnWrite implements Wikitty {

    private static final long serialVersionUID = 1L;

    private static final Log log = LogFactory.getLog(WikittyCopyOnWrite.class);

    protected Wikitty target;

    protected boolean targetIsACopy = false;

    /** only WikittyService with cache should create instances */
    protected WikittyCopyOnWrite(Wikitty target) {
        this.target = target;
    }
    
    /** replace {@link #target} with a clone
     *
     * this method must be called to prevent any modification on target
     */
    protected void substituteTargetWithCopy() {
        // test make a the copy once
        if (! targetIsACopy) {
            try {
                target = target.clone();
                if (log.isTraceEnabled()) {
                    log.trace(this + " now has for target " + target);
                }
            } catch (CloneNotSupportedException e) {
                log.error("unable to clone wikitty " + target, e);
            }
            targetIsACopy = true;            
        }
    }

    public boolean equals(Object obj) {
        // FIXME 20100804 bleny side-effect ?
        return target.equals(obj);
    }

    public int hashCode() {
        // FIXME 20100804 bleny side-effect ?
        return target.hashCode();
    }

    public String toString() {
        return "[" + WikittyCopyOnWrite.class.getName() + "]" + target.toString();
    }

    @Override
    public Wikitty clone() throws CloneNotSupportedException {
        // return a clone of the target
        return target.clone();
    }

    /* ** below are only delegation code with copy-on-write */

    public void addPropertyChangeListener(PropertyChangeListener listener) {
        target.addPropertyChangeListener(listener);
    }

    public void removePropertyChangeListener(PropertyChangeListener listener) {
        target.removePropertyChangeListener(listener);
    }

    public void addPropertyChangeListener(String propertyName,
            PropertyChangeListener listener) {
        target.addPropertyChangeListener(propertyName, listener);
    }

    public void removePropertyChangeListener(String propertyName,
            PropertyChangeListener listener) {
        target.removePropertyChangeListener(propertyName, listener);
    }

    public String getId() {
        return target.getId();
    }

    public boolean isDeleted() {
        return target.isDeleted();
    }

    public Date getDeleteDate() {
        return target.getDeleteDate();
    }

    public void setDeleteDate(Date delete) {
        target.setDeleteDate(delete);
    }

    public void addExtension(WikittyExtension ext) {
        substituteTargetWithCopy();
        target.addExtension(ext);
    }

    public void addExtension(List<WikittyExtension> exts) {
        substituteTargetWithCopy();
        target.addExtension(exts);
    }

    public boolean hasExtension(String extName) {
        return target.hasExtension(extName);
    }

    public boolean hasField(String extName, String fieldName) {
        return target.hasField(extName, fieldName);
    }

    public WikittyExtension getExtension(String ext) {
        return target.getExtension(ext);
    }

    public Collection<String> getExtensionNames() {
        return target.getExtensionNames();
    }

    public Collection<WikittyExtension> getExtensions() {
        return target.getExtensions();
    }

    public Collection<WikittyExtension> getExtensionDependencies(String ext,
            boolean recursively) {
        return target.getExtensionDependencies(ext, recursively);
    }

    public FieldType getFieldType(String fqfieldName) {
        return target.getFieldType(fqfieldName);
    }

    public void setField(String ext, String fieldName, Object value) {
        substituteTargetWithCopy();
        target.setField(ext, fieldName, value);
    }

    public Object getFieldAsObject(String ext, String fieldName) {
        return target.getFieldAsObject(ext, fieldName);
    }

    public boolean getFieldAsBoolean(String ext, String fieldName) {
        return target.getFieldAsBoolean(ext, fieldName);
    }

    public BigDecimal getFieldAsBigDecimal(String ext, String fieldName) {
        return target.getFieldAsBigDecimal(ext, fieldName);
    }

    public int getFieldAsInt(String ext, String fieldName) {
        return target.getFieldAsInt(ext, fieldName);
    }

    public long getFieldAsLong(String ext, String fieldName) {
        return target.getFieldAsLong(ext, fieldName);
    }

    public float getFieldAsFloat(String ext, String fieldName) {
        return target.getFieldAsFloat(ext, fieldName);
    }

    public double getFieldAsDouble(String ext, String fieldName) {
        return target.getFieldAsDouble(ext, fieldName);
    }

    public String getFieldAsString(String ext, String fieldName) {
        return target.getFieldAsString(ext, fieldName);
    }

    public Date getFieldAsDate(String ext, String fieldName) {
        return target.getFieldAsDate(ext, fieldName);
    }

    public String getFieldAsWikitty(String ext, String fieldName) {
        return target.getFieldAsWikitty(ext, fieldName);
    }

    public <E> List<E> getFieldAsList(String ext, String fieldName,
            Class<E> clazz) {
        return target.getFieldAsList(ext, fieldName, clazz);
    }

    public <E> Set<E> getFieldAsSet(String ext, String fieldName, Class<E> clazz) {
        return target.getFieldAsSet(ext, fieldName, clazz);
    }

    public void addToField(String ext, String fieldName, Object value) {
        substituteTargetWithCopy();
        target.addToField(ext, fieldName, value);
    }

    public void removeFromField(String ext, String fieldName, Object value) {
        substituteTargetWithCopy();
        target.removeFromField(ext, fieldName, value);
    }

    public void clearField(String ext, String fieldName) {
        substituteTargetWithCopy();
        target.clearField(ext, fieldName);
    }

    public Set<String> fieldNames() {
        return target.fieldNames();
    }

    public Object getFqField(String fqFieldName) {
        return target.getFqField(fqFieldName);
    }

    public String getVersion() {
        return target.getVersion();
    }

    public void setVersion(String version) {
        substituteTargetWithCopy();
        target.setVersion(version);
    }

    public void clearDirty() {
        target.clearDirty();
    }

    public void setFqField(String fieldName, Object value) {
        target.setFqField(fieldName, value);
    }

    public boolean isEmpty() {
        return target.isEmpty();
    }
    
}

==> wikitty-api-2.2-sources.jar: org/nuiton/wikitty/WikittyCopyOnWrite.java <==
package org.nuiton.wikitty;

import java.beans.PropertyChangeListener;
import java.math.BigDecimal;
import java.util.Collection;
import java.util.Date;
import java.util.List;
import java.util.Set;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/** this class wrap a wikitty in a proxy that copy the wikitty before applying any change
 *
 * It is used for caching purpose. This class own a reference to an actual
 * wikitty, multiple instance of this class can own some references to a same
 * actual wikitty. To prevent the target to be modified, a copy is created
 * to prevent side-effect.
 * 
 * So, when a wikitty restored from cache is modified, a copy is modified, so
 * if change are cancelled, next restore will restore the original and not
 * the modified version (until the modified version is stored).
 * 
 * used in {@link WikittyServiceCached}
 */
public class WikittyCopyOnWrite implements Wikitty {

    private static final long serialVersionUID = 1L;

    private static final Log log = LogFactory.getLog(WikittyCopyOnWrite.class);

    protected Wikitty target;

    protected boolean targetIsACopy = false;

    /** only WikittyService with cache should create instances */
    protected WikittyCopyOnWrite(Wikitty target) {
        this.target = target;
    }
    
    /** replace {@link #target} with a clone
     *
     * this method must be called to prevent any modification on target
     */
    protected void substituteTargetWithCopy() {
        // test make a the copy once
        if (! targetIsACopy) {
            try {
                target = target.clone();
                if (log.isTraceEnabled()) {
                    log.trace(this + " now has for target " + target);
                }
            } catch (CloneNotSupportedException e) {
                log.error("unable to clone wikitty " + target, e);
            }
            targetIsACopy = true;            
        }
    }

    public boolean equals(Object obj) {
        // FIXME 20100804 bleny side-effect ?
        return target.equals(obj);
    }

    public int hashCode() {
        // FIXME 20100804 bleny side-effect ?
        return target.hashCode();
    }

    public String toString() {
        return "[" + WikittyCopyOnWrite.class.getName() + "]" + target.toString();
    }

    @Override
    public Wikitty clone() throws CloneNotSupportedException {
        // return a clone of the target
        return target.clone();
    }

    /* ** below are only delegation code with copy-on-write */

    public void addPropertyChangeListener(PropertyChangeListener listener) {
        target.addPropertyChangeListener(listener);
    }

    public void removePropertyChangeListener(PropertyChangeListener listener) {
        target.removePropertyChangeListener(listener);
    }

    public void addPropertyChangeListener(String propertyName,
            PropertyChangeListener listener) {
        target.addPropertyChangeListener(propertyName, listener);
    }

    public void removePropertyChangeListener(String propertyName,
            PropertyChangeListener listener) {
        target.removePropertyChangeListener(propertyName, listener);
    }

    public String getId() {
        return target.getId();
    }

    public boolean isDeleted() {
        return target.isDeleted();
    }

    public Date getDeleteDate() {
        return target.getDeleteDate();
    }

    public void setDeleteDate(Date delete) {
        target.setDeleteDate(delete);
    }

    public void addExtension(WikittyExtension ext) {
        substituteTargetWithCopy();
        target.addExtension(ext);
    }

    public void addExtension(List<WikittyExtension> exts) {
        substituteTargetWithCopy();
        target.addExtension(exts);
    }

    public boolean hasExtension(String extName) {
        return target.hasExtension(extName);
    }

    public boolean hasField(String extName, String fieldName) {
        return target.hasField(extName, fieldName);
    }

    public WikittyExtension getExtension(String ext) {
        return target.getExtension(ext);
    }

    public Collection<String> getExtensionNames() {
        return target.getExtensionNames();
    }

    public Collection<WikittyExtension> getExtensions() {
        return target.getExtensions();
    }

    public Collection<WikittyExtension> getExtensionDependencies(String ext,
            boolean recursively) {
        return target.getExtensionDependencies(ext, recursively);
    }

    public FieldType getFieldType(String fqfieldName) {
        return target.getFieldType(fqfieldName);
    }

    public void setField(String ext, String fieldName, Object value) {
        substituteTargetWithCopy();
        target.setField(ext, fieldName, value);
    }

    public Object getFieldAsObject(String ext, String fieldName) {
        return target.getFieldAsObject(ext, fieldName);
    }

    public boolean getFieldAsBoolean(String ext, String fieldName) {
        return target.getFieldAsBoolean(ext, fieldName);
    }

    public BigDecimal getFieldAsBigDecimal(String ext, String fieldName) {
        return target.getFieldAsBigDecimal(ext, fieldName);
    }

    public int getFieldAsInt(String ext, String fieldName) {
        return target.getFieldAsInt(ext, fieldName);
    }

    public long getFieldAsLong(String ext, String fieldName) {
        return target.getFieldAsLong(ext, fieldName);
    }

    public float getFieldAsFloat(String ext, String fieldName) {
        return target.getFieldAsFloat(ext, fieldName);
    }

    public double getFieldAsDouble(String ext, String fieldName) {
        return target.getFieldAsDouble(ext, fieldName);
    }

    public String getFieldAsString(String ext, String fieldName) {
        return target.getFieldAsString(ext, fieldName);
    }

    public Date getFieldAsDate(String ext, String fieldName) {
        return target.getFieldAsDate(ext, fieldName);
    }

    public String getFieldAsWikitty(String ext, String fieldName) {
        return target.getFieldAsWikitty(ext, fieldName);
    }

    public <E> List<E> getFieldAsList(String ext, String fieldName,
            Class<E> clazz) {
        return target.getFieldAsList(ext, fieldName, clazz);
    }

    public <E> Set<E> getFieldAsSet(String ext, String fieldName, Class<E> clazz) {
        return target.getFieldAsSet(ext, fieldName, clazz);
    }

    public void addToField(String ext, String fieldName, Object value) {
        substituteTargetWithCopy();
        target.addToField(ext, fieldName, value);
    }

    public void removeFromField(String ext, String fieldName, Object value) {
        substituteTargetWithCopy();
        target.removeFromField(ext, fieldName, value);
    }

    public void clearField(String ext, String fieldName) {
        substituteTargetWithCopy();
        target.clearField(ext, fieldName);
    }

    /**
     * @see org.nuiton.wikitty.Wikitty#addToField(String, Object)
     */
    @Override
    public void addToField(String fqFieldName, Object value) {
        substituteTargetWithCopy();
        target.addToField(fqFieldName, value);
    }

    /**
     * @see org.nuiton.wikitty.Wikitty#removeFromField(String, Object)
     */
    @Override
    public void removeFromField(String fqFieldName, Object value) {
        substituteTargetWithCopy();
        target.removeFromField(fqFieldName, value);
    }

    /**
     * @see org.nuiton.wikitty.Wikitty#clearField(String)
     */
    @Override
    public void clearField(String fqFieldName) {
        substituteTargetWithCopy();
        target.clearField(fqFieldName);
    }

    public Set<String> fieldNames() {
        return target.fieldNames();
    }

    public Object getFqField(String fqFieldName) {
        return target.getFqField(fqFieldName);
    }

    public String getVersion() {
        return target.getVersion();
    }

    public void setVersion(String version) {
        substituteTargetWithCopy();
        target.setVersion(version);
    }

    /**
     * @see org.nuiton.wikitty.Wikitty#getDirty()
     */
    public Set<String> getDirty() {
        return target.getDirty();
    }

    /**
     * @see org.nuiton.wikitty.Wikitty#clearDirty()
     */
    public void clearDirty() {
        substituteTargetWithCopy();
        target.clearDirty();
    }

    /**
     * @see org.nuiton.wikitty.Wikitty#setFqField(String, Object)
     */
    public void setFqField(String fieldName, Object value) {
        substituteTargetWithCopy();
        target.setFqField(fieldName, value);
    }

    public boolean isEmpty() {
        return target.isEmpty();
    }

    /**
     * @see org.nuiton.wikitty.Wikitty#hasMetaExtension(String, String)
     */
    @Override
    public boolean hasMetaExtension(String metaExtensionName,
                                    String extensionName) {
        return target.hasMetaExtension(metaExtensionName, extensionName);
    }
    
    /**
     * @see org.nuiton.wikitty.Wikitty#addMetaExtension(WikittyExtension, WikittyExtension)
     */
    @Override
    public void addMetaExtension(WikittyExtension metaExtension,
            WikittyExtension extension) {
        substituteTargetWithCopy();
        target.addMetaExtension(metaExtension, extension);
    }

    /**
     * @see org.nuiton.wikitty.Wikitty#addMetaExtension(WikittyExtension, String)
     */
    @Override
    public void addMetaExtension(WikittyExtension metaExtension,
            String extensionFqn) {
        substituteTargetWithCopy();
        target.addMetaExtension(metaExtension, extensionFqn);
    }
    
}
